"""BSMP serial communications classes."""
import abc
import struct as _struct
import typing
from threading import Lock as _Lock

from .exceptions import SerialErrCheckSum as _SerialErrCheckSum
from .exceptions import SerialErrEmpty as _SerialErrEmpty
from .exceptions import SerialErrMsgShort as _SerialErrMsgShort
from .exceptions import SerialErrPckgLen as _SerialErrPckgLen


class IOInterface(metaclass=abc.ABCMeta):
    """Base class for I/O"""

    @abc.abstractmethod
    def open(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def UART_read(self) -> typing.List[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def UART_write(self, stream, timeout: float) -> typing.Optional[typing.Any]:
        raise NotImplementedError

    @abc.abstractmethod
    def UART_request(self, stream, timeout: float) -> typing.Optional[typing.List[str]]:
        raise NotImplementedError


class Message:
    """BSMP Message.

    Command: command id; 1 byte;
    Load Size: payload size in bytes; 2 bytes (big endian);
    Load: 0..65535 bytes.
    """

    # TODO: indicate somehow that stream is a char stream.

    # Constructors
    def __init__(self, stream: typing.List[str]):
        """Build a BSMP message."""
        if len(stream) < 3:
            raise _SerialErrMsgShort(
                f"BSMP Message too short (stream: {stream}).")
        self._stream: typing.List[str] = stream
        self._cmd: int = ord(stream[0])

    def __eq__(self, other) -> bool:
        """Compare messages."""
        return self._stream == other.stream

    @classmethod
    def message(
        cls,
        cmd: int,
        payload: typing.Optional[typing.List[str]] = None
    ):
        """Build a Message object from a byte stream."""
        if payload and not isinstance(payload, list):
            # TODO: should be create serial exceptions here too?
            raise TypeError("Load must be a list.")
        if payload and len(payload) > 65535:
            # TODO: should be create serial exceptions here too?
            raise ValueError("Load must be smaller than 65535.")

        stream: typing.List[str] = []

        if not payload:
            payload = []
        # Append cmd
        stream.append(chr(cmd))
        # Append size
        stream.extend(list(map(chr, (_struct.pack('>H', len(payload))))))
        # Append payload
        stream.extend(payload)
        return cls(stream)

    # API
    @property
    def stream(self) -> typing.List[str]:
        """Return stream."""
        return self._stream

    @property
    def cmd(self) -> int:
        """Command ID."""
        return self._cmd

    @property
    def size(self) -> int:
        """Load size."""
        return _struct.unpack('>H', bytes(map(ord, self._stream[1:3])))[0]

    @property
    def payload(self) -> typing.List[str]:
        """Message payload."""
        return self._stream[3:]


class Package:
    """BSMP Package.

    Address: 1 byte, 0..31
    Message: no limit
    Checksum: package checksum
    """

    def __init__(
        self,
        stream: typing.List[str]
    ):
        """Build a BSMP package."""
        if len(stream) < 5:
            raise _SerialErrPckgLen(
                f"BSMP Package too short (stream: {stream}).")
        if not Package.verify_checksum(stream):
            raise _SerialErrCheckSum(
                f"Inconsistent message. Checksum does not check (stream: {stream}).")

        self._stream: typing.List[str] = stream
        self._address: int = ord(stream[0])  # 0 to 31
        self._message: Message = Message(stream[1:-1])
        self._checksum: int = ord(stream[-1])

    @classmethod
    def package(cls, address: int, message: Message):
        """Build a Package object from a byte stream."""
        # Return new package
        stream: typing.List[str] = []
        stream.append(chr(address))
        stream.extend(message.stream)
        chksum = cls.calc_checksum(stream)
        stream.append(chr(chksum))
        return cls(stream)

    # API
    @property
    def stream(self) -> typing.List[str]:
        """Stream."""
        return self._stream

    @property
    def address(self) -> int:
        """Receiver node serial address."""
        return self._address

    @property
    def message(self) -> Message:
        """Return package message."""
        return self._message

    @property
    def checksum(self) -> int:
        """Return package checksum."""
        return self._checksum

    @staticmethod
    def calc_checksum(stream) -> int:
        """Return stream checksum."""
        streambytes = [ord(s) for s in stream]
        counter = sum(streambytes)
        counter = (counter & 0xFF)
        counter = (256 - counter) & 0xFF
        return counter

    @staticmethod
    def verify_checksum(stream: typing.List[str]) -> bool:
        """Verify stream checksum."""
        streambytes = [ord(s) for s in stream[:-1]]
        counter = sum(streambytes)
        counter = (counter & 0xFF)
        counter = (256 - counter) & 0xFF
        if stream[len(stream) - 1] == chr(counter):
            return True
        else:
            return False


class Channel:
    """BSMP Channel.

    The channel is defined by an IOInterface object and recipient
    address.
    """

    # TODO: should we remove use of default timeout values. It is dangerous!
    LOCK: typing.Optional[_Lock] = None

    def __init__(self, iointerf: IOInterface, address: int):
        """Set channel."""
        self._iointerf: IOInterface = iointerf  # IOInterface object to communicate with bsmp device
        self._address: int = address  # address of recipient device.
        self._size_counter: int = 0  # stream size counter [bytes]

    @property
    def iointerf(self) -> IOInterface:
        """Return IOInterface serial communication object."""
        return self._iointerf

    @property
    def address(self) -> int:
        """Return attached bsmp device id."""
        return self._address

    @property
    def size_counter(self) -> int:
        """Return stream size of last request."""
        return self._size_counter

    def size_counter_reset(self) -> None:
        """Reset stream size counter."""
        self._size_counter = 0

    def read(self) -> Message:
        """Read from serial."""
        resp = self.iointerf.UART_read()
        if not resp:
            raise _SerialErrEmpty("Serial read returned empty!")
        package = Package(resp)
        self._size_counter += len(package.stream)
        return package.message

    def write(self, message: Message, timeout: float = 100):
        """Write to serial. :param timeout [ms]"""
        stream = Package.package(self._address, message).stream
        response = self.iointerf.UART_write(stream, timeout=timeout)
        self._size_counter += len(stream)
        return response

    def request_(self, message: Message, timeout: float = 100) -> Message:
        """:param timeout [ms]"""
        stream = Package.package(self._address, message).stream

        if Channel.LOCK is None:
            response = self.iointerf.UART_request(stream, timeout=timeout)
        else:
            with Channel.LOCK:
                response = self.iointerf.UART_request(stream, timeout=timeout)

        self._size_counter += len(stream)

        if not response:
            raise _SerialErrEmpty("Serial read returned empty!")

        package = Package(response)
        self._size_counter += len(package.stream)
        return package.message

    def request(self, message: Message, timeout: float = 100, read_flag: bool = True) -> typing.Optional[Message]:
        """Write and wait for response. :param timeout [ms]"""
        response: typing.Optional[Message] = None

        if read_flag:
            response = self.request_(message, timeout)
        else:
            self.write(message, timeout)

        return response

    @staticmethod
    def create_lock() -> None:
        """."""
        Channel.LOCK = _Lock()



"""AS PowerState On configuration."""
from copy import deepcopy as _dcopy


_on = 1


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict),
        'check': False,
    }
    return _dict


# When using this type of configuration to set the machine,
# the list of PVs should be processed in the same order they are stored
# in the configuration. The second numeric parameter in the pair is the
# delay [s] the client should wait before setting the next PV.


_pvs_li_ps = [
    ['LI-01:PS-LensRev:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Lens-1:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Lens-2:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Lens-3:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Lens-4:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-CV-1:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-CH-1:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-CV-2:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-CH-2:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-CV-3:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-CH-3:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-CV-4:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-CH-4:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-CV-5:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-CH-5:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-CV-6:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-CH-6:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-CV-7:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-CH-7:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Slnd-1:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Slnd-2:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Slnd-3:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Slnd-4:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Slnd-5:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Slnd-6:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Slnd-7:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Slnd-8:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Slnd-9:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Slnd-10:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Slnd-11:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Slnd-12:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Slnd-13:PwrState-Sel', _on, 0.0],
    ['LI-Fam:PS-Slnd-14:PwrState-Sel', _on, 0.0],
    ['LI-Fam:PS-Slnd-15:PwrState-Sel', _on, 0.0],
    ['LI-Fam:PS-Slnd-16:PwrState-Sel', _on, 0.0],
    ['LI-Fam:PS-Slnd-17:PwrState-Sel', _on, 0.0],
    ['LI-Fam:PS-Slnd-18:PwrState-Sel', _on, 0.0],
    ['LI-Fam:PS-Slnd-19:PwrState-Sel', _on, 0.0],
    ['LI-Fam:PS-Slnd-20:PwrState-Sel', _on, 0.0],
    ['LI-Fam:PS-Slnd-21:PwrState-Sel', _on, 0.0],
    ['LI-Fam:PS-QF1:PwrState-Sel', _on, 0.0],
    ['LI-Fam:PS-QF2:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-QF3:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-QD1:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-QD2:PwrState-Sel', _on, 0.0],
    ['LI-01:PS-Spect:PwrState-Sel', _on, 0.0],
    ]

_pvs_tb_ps = [
    ['TB-Fam:PS-B:PwrState-Sel', _on, 0.0],
    ['TB-01:PS-QD1:PwrState-Sel', _on, 0.0],
    ['TB-01:PS-QF1:PwrState-Sel', _on, 0.0],
    ['TB-02:PS-QD2A:PwrState-Sel', _on, 0.0],
    ['TB-02:PS-QF2A:PwrState-Sel', _on, 0.0],
    ['TB-02:PS-QF2B:PwrState-Sel', _on, 0.0],
    ['TB-02:PS-QD2B:PwrState-Sel', _on, 0.0],
    ['TB-03:PS-QF3:PwrState-Sel', _on, 0.0],
    ['TB-03:PS-QD3:PwrState-Sel', _on, 0.0],
    ['TB-04:PS-QF4:PwrState-Sel', _on, 0.0],
    ['TB-04:PS-QD4:PwrState-Sel', _on, 0.0],
    ['TB-01:PS-CH-1:PwrState-Sel', _on, 0.0],
    ['TB-01:PS-CV-1:PwrState-Sel', _on, 0.0],
    ['TB-01:PS-CH-2:PwrState-Sel', _on, 0.0],
    ['TB-01:PS-CV-2:PwrState-Sel', _on, 0.0],
    ['TB-02:PS-CH-1:PwrState-Sel', _on, 0.0],
    ['TB-02:PS-CV-1:PwrState-Sel', _on, 0.0],
    ['TB-02:PS-CH-2:PwrState-Sel', _on, 0.0],
    ['TB-02:PS-CV-2:PwrState-Sel', _on, 0.0],
    ['TB-04:PS-CH-1:PwrState-Sel', _on, 0.0],
    ['TB-04:PS-CV-1:PwrState-Sel', _on, 0.0],
    ['TB-04:PS-CH-2:PwrState-Sel', _on, 0.0],
    ['TB-04:PS-CV-2:PwrState-Sel', _on, 0.0],
    ]

_pvs_pu = [
    ['TB-04:PU-InjSept:PwrState-Sel', 0, 0.0],
    ['BO-01D:PU-InjKckr:PwrState-Sel', 0, 0.0],
    ]

_pvs_bo_ps = [
    ['BO-Fam:PS-B-1:PwrState-Sel', _on, 0.0],
    ['BO-Fam:PS-B-2:PwrState-Sel', _on, 0.0],
    ['BO-Fam:PS-QD:PwrState-Sel', _on, 0.0],
    ['BO-Fam:PS-QF:PwrState-Sel', _on, 0.0],
    ['BO-02D:PS-QS:PwrState-Sel', _on, 0.0],
    ['BO-Fam:PS-SD:PwrState-Sel', _on, 0.0],
    ['BO-Fam:PS-SF:PwrState-Sel', _on, 0.0],
    ['BO-01U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-03U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-05U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-07U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-09U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-11U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-13U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-15U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-17U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-19U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-21U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-23U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-25U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-27U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-29U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-31U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-33U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-35U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-37U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-39U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-41U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-43U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-45U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-47U:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-49D:PS-CH:PwrState-Sel', _on, 0.0],
    ['BO-01U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-03U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-05U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-07U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-09U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-11U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-13U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-15U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-17U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-19U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-21U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-23U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-25U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-27U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-29U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-31U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-33U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-35U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-37U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-39U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-41U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-43U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-45U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-47U:PS-CV:PwrState-Sel', _on, 0.0],
    ['BO-49U:PS-CV:PwrState-Sel', _on, 0.0],
    ]

_template_dict = {
    'pvs':
        _pvs_li_ps + _pvs_tb_ps + _pvs_pu + _pvs_bo_ps
    }



"""."""

from .device import Device as _Device


class LIEnergy(_Device):
    """Linac Energy Measurement Device."""

    class DEVICES:
        """Devices names."""

        LI = 'LI-Glob:AP-MeasEnergy'
        ALL = (LI, )

    PROPERTIES_DEFAULT = (
        'Dispersion-SP', 'Dispersion-RB',
        'IntDipole-Mon', 'Energy-Mon', 'Spread-Mon',
        'MeasureCtrl-Sel', 'MeasureCtrl-Sts')

    def __init__(self, devname=None, props2init='all'):
        """."""
        # check if device exists
        devname = devname if devname else self.DEVICES.LI
        if devname not in LIEnergy.DEVICES.ALL:
            raise NotImplementedError(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def dispersion(self):
        """."""
        return self['Dispersion-RB']

    @dispersion.setter
    def dispersion(self, val):
        self['Dispersion-SP'] = val

    @property
    def intdipole(self):
        """."""
        return self['IntDipole-Mon']

    @property
    def energy(self):
        """."""
        return self['Energy-Mon']

    @property
    def energy_spread(self):
        """."""
        return self['Spread-Mon']

    @property
    def is_measuring(self):
        """."""
        return self['MeasureCtrl-Sts']

    def cmd_turn_on_measurement(self, timeout=10):
        """."""
        self['MeasureCtrl-Sel'] = 1
        return self._wait('MeasureCtrl-Sts', 1, timeout=timeout)

    def cmd_turn_off_measurement(self, timeout=10):
        """."""
        self['MeasureCtrl-Sel'] = 0
        return self._wait('MeasureCtrl-Sts', 0, timeout=timeout)



"""Define Screen device high level PVs database.

This is not a primary source database. Primary sources can be found in:
 - Screen IOC: https://github.com/lnls-dig/screen-epics-ioc
 - Camera IOC: https://github.com/lnls-dig/basler-acA1300-75gm-epics-ioc
 - Motor IOC: https://github.com/lnls-dig/galil-dmc30017-epics-ioc
"""

from ... import csdev as _csdev


# --- Enumeration Types ---

class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    SCRNTYPESEL = ('None', 'Calibration', 'Fluorescent')
    SCRNTYPESTS = ('None', 'Calibration', 'Fluorescent', 'Unknown')
    CAMACQMODE = ('AUTO', 'TRIG')
    CAMEXPOSUREMODE = ('TIMED', 'TRIGWIDTH')
    CAMLASTERR = ('NoError', 'Overtrigger', 'Userset', 'InvalidParameter',
                  'OverTemperature', 'PowerFailure', 'InsufficientTrig',
                  'UserDefPixFailur')
    CAMTEMPSTATE = ('Ok', 'Critical', 'Error')


_et = ETypes  # syntactic sugar


# --- Const class ---

class Const(_csdev.Const):
    """Const class defining Screens constants and Enum types."""

    CamAcqMode = _csdev.Const.register('CamAcqMode', _et.CAMACQMODE)
    CamExposureMode = _csdev.Const.register('CamExposureMode',
                                            _et.CAMEXPOSUREMODE)
    CamLastErr = _csdev.Const.register('CamLastErr', _et.CAMLASTERR)
    CamTempState = _csdev.Const.register('CamTempState', _et.CAMTEMPSTATE)


_c = Const  # syntactic sugar


# --- Database ---

def get_scrn_database():
    """Return Scrn device database."""
    pvs_database = {
        # Low level devides prefixes
        'MtrCtrlPrefix-Cte': {'type': 'string'},
        'CamPrefix-Cte':     {'type': 'string'},

        # PV to select screen type
        'ScrnType-Sel': {'type': 'enum', 'enums': _et.SCRNTYPESEL, 'value': 0},
        'ScrnType-Sts': {'type': 'enum', 'enums': _et.SCRNTYPESTS, 'value': 0},

        # Camera settings PVs
        'CamEnbl-Sel':         {'type': 'enum', 'enums': _et.OFF_ON,
                                'value': 0},
        'CamEnbl-Sts':         {'type': 'enum', 'enums': _et.OFF_ON,
                                'value': 0},
        'CamAcqMode-Sel':      {'type': 'enum', 'enums': _et.CAMACQMODE,
                                'value': 0},
        'CamAcqMode-Sts':      {'type': 'enum', 'enums': _et.CAMACQMODE,
                                'value': 0},
        'CamAcqPeriod-SP':     {'type': 'float', 'unit': 's', 'prec': 6,
                                'value': 0.2, 'hihi': 10, 'high': 10,
                                'hilim': 10, 'lolim': 0.0124, 'low': 0.0124,
                                'lolo': 0.0124},
        'CamAcqPeriod-RB':     {'type': 'float', 'unit': 's', 'prec': 6,
                                'value': 0.2, 'hihi': 10, 'high': 10,
                                'hilim': 10, 'lolim': 0.0124, 'low': 0.0124,
                                'lolo': 0.0124},
        'CamExposureMode-Sel': {'type': 'enum', 'enums': _et.CAMEXPOSUREMODE,
                                'value': 0},
        'CamExposureMode-Sts': {'type': 'enum', 'enums': _et.CAMEXPOSUREMODE,
                                'value': 0},
        'CamExposureTime-SP':  {'type': 'int', 'unit': 'us', 'value': 5000,
                                'hihi': 1000000, 'high': 1000000,
                                'hilim': 1000000, 'lolim': 1, 'low': 1,
                                'lolo': 1},
        'CamExposureTime-RB':  {'type': 'int', 'unit': 'us', 'value': 5000,
                                'hihi': 1000000, 'high': 1000000,
                                'hilim': 1000000, 'lolim': 1, 'low': 1,
                                'lolo': 1},
        'CamGain-SP':          {'type': 'float', 'unit': 'dB', 'prec': 6,
                                'hihi': 20, 'high': 20, 'hilim': 20,
                                'lolim': 0, 'low': 0, 'lolo': 0},
        'CamGain-RB':          {'type': 'float', 'unit': 'dB', 'prec': 6,
                                'hihi': 20, 'high': 20, 'hilim': 20,
                                'lolim': 0, 'low': 0, 'lolo': 0},
        'CamAutoGain-Cmd':     {'type': 'int', 'value': 0},
        'CamBlackLevel-SP':    {'type': 'int', 'unit': 'gray value',
                                'hihi': 64, 'high': 64, 'hilim': 64,
                                'lolim': 0, 'low': 0, 'lolo': 0},
        'CamBlackLevel-RB':    {'type': 'int', 'unit': 'gray value',
                                'hihi': 64, 'high': 64, 'hilim': 64,
                                'lolim': 0, 'low': 0, 'lolo': 0},
        'CamLastErr-Mon':      {'type': 'enum', 'enums': _et.CAMLASTERR},
        'CamClearLastErr-Cmd': {'type': 'int', 'value': 0},
        'CamTempState-Mon':    {'type': 'enum', 'enums': _et.CAMTEMPSTATE},
        'CamTemp-Mon':         {'type': 'float', 'unit': '°C', 'prec': 3},

        # Camera image PVs
        'ImgData-Mon':      {'type': 'int'},
        'ImgMaxWidth-Cte':  {'type': 'int', 'unit': 'pixels'},
        'ImgMaxHeight-Cte': {'type': 'int', 'unit': 'pixels'},

        # Camera ROI settings PVs
        'ImgROIOffsetX-SP':      {'type': 'int', 'unit': 'pixels'},
        'ImgROIOffsetX-RB':      {'type': 'int', 'unit': 'pixels'},
        'ImgROIOffsetY-SP':      {'type': 'int', 'unit': 'pixels'},
        'ImgROIOffsetY-RB':      {'type': 'int', 'unit': 'pixels'},
        'ImgROIWidth-SP':        {'type': 'int', 'unit': 'pixels'},
        'ImgROIWidth-RB':        {'type': 'int', 'unit': 'pixels'},
        'ImgROIHeight-SP':       {'type': 'int', 'unit': 'pixels'},
        'ImgROIHeight-RB':       {'type': 'int', 'unit': 'pixels'},
        'ImgROIAutoCenterX-Sel': {'type': 'int', 'value': 0},
        'ImgROIAutoCenterX-Sts': {'type': 'int', 'value': 0},
        'ImgROIAutoCenterY-Sel': {'type': 'int', 'value': 0},
        'ImgROIAutoCenterY-Sts': {'type': 'int', 'value': 0},

        # Camera image statistics PVs
        'CenterXNDStatsRaw-Mon': {'type': 'int', 'unit': 'pixels'},
        'CenterYNDStatsRaw-Mon': {'type': 'int', 'unit': 'pixels'},
        'SigmaXNDStatsRaw-Mon':  {'type': 'int', 'unit': 'pixels'},
        'SigmaYNDStatsRaw-Mon':  {'type': 'int', 'unit': 'pixels'},
        'SigmaXYNDStatsRaw-Mon': {'type': 'int', 'unit': 'pixels'},
        'CenterXDimFeiRaw-Mon':  {'type': 'int', 'unit': 'pixels'},
        'CenterYDimFeiRaw-Mon':  {'type': 'int', 'unit': 'pixels'},
        'SigmaXDimFeiRaw-Mon':   {'type': 'int', 'unit': 'pixels'},
        'SigmaYDimFeiRaw-Mon':   {'type': 'int', 'unit': 'pixels'},
        'ThetaNDStats-Mon':      {'type': 'float', 'unit': 'rad', 'prec': 6},
        'ThetaDimFei-Mon':       {'type': 'float', 'unit': 'rad', 'prec': 6},

        # Camera image statistics PVs in mm
        'CenterXNDStats-Mon': {'type': 'float', 'unit': 'mm', 'prec': 6},
        'CenterYNDStats-Mon': {'type': 'float', 'unit': 'mm', 'prec': 6},
        'SigmaXNDStats-Mon':  {'type': 'float', 'unit': 'mm', 'prec': 6},
        'SigmaYNDStats-Mon':  {'type': 'float', 'unit': 'mm', 'prec': 6},
        'SigmaXYNDStats-Mon': {'type': 'float', 'unit': 'mm', 'prec': 6},
        'CenterXDimFei-Mon':  {'type': 'float', 'unit': 'mm', 'prec': 6},
        'CenterYDimFei-Mon':  {'type': 'float', 'unit': 'mm', 'prec': 6},
        'SigmaXDimFei-Mon':   {'type': 'float', 'unit': 'mm', 'prec': 6},
        'SigmaYDimFei-Mon':   {'type': 'float', 'unit': 'mm', 'prec': 6},

        # Screens position calibration PVs
        'FluorScrnPos-SP':     {'type': 'float', 'unit': 'mm', 'prec': 5,
                                'hihi': 1000, 'high': 1000, 'hilim': 1000,
                                'lolim': -1000, 'low': -1000, 'lolo': -1000},
        'FluorScrnPos-RB':     {'type': 'float', 'unit': 'mm', 'prec': 5,
                                'hihi': 1000, 'high': 1000, 'hilim': 1000,
                                'lolim': -1000, 'low': -1000, 'lolo': -1000},
        'GetFluorScrnPos-Cmd': {'type': 'int', 'value': 0},
        'CalScrnPos-SP':       {'type': 'float', 'unit': 'mm', 'prec': 5,
                                'hihi': 1000, 'high': 1000, 'hilim': 1000,
                                'lolim': -1000, 'low': -1000, 'lolo': -1000},
        'CalScrnPos-RB':       {'type': 'float', 'unit': 'mm', 'prec': 5,
                                'hihi': 1000, 'high': 1000, 'hilim': 1000,
                                'lolim': -1000, 'low': -1000, 'lolo': -1000},
        'GetCalScrnPos-Cmd':   {'type': 'int', 'value': 0},
        'NoneScrnPos-SP':      {'type': 'float', 'unit': 'mm', 'prec': 5,
                                'hihi': 1000, 'high': 1000, 'hilim': 1000,
                                'lolim': -1000, 'low': -1000, 'lolo': -1000},
        'NoneScrnPos-RB':      {'type': 'float', 'unit': 'mm', 'prec': 5,
                                'hihi': 1000, 'high': 1000, 'hilim': 1000,
                                'lolim': -1000, 'low': -1000, 'lolo': -1000},
        'GetNoneScrnPos-Cmd':  {'type': 'int', 'value': 0},
        'AcceptedErr-SP':      {'type': 'float', 'unit': 'mm', 'prec': 5,
                                'hihi': 1000, 'high': 1000, 'hilim': 1000,
                                'lolim': -1000, 'low': -1000, 'lolo': -1000},
        'AcceptedErr-RB':      {'type': 'float', 'unit': 'mm', 'prec': 5,
                                'hihi': 1000, 'high': 1000, 'hilim': 1000,
                                'lolim': -1000, 'low': -1000, 'lolo': -1000},

        # LED brightness calibration PVs
        'EnblLED-Sel':          {'type': 'enum', 'enums': _et.OFF_ON,
                                 'value': 0},
        'EnblLED-Sts':          {'type': 'enum', 'enums': _et.OFF_ON,
                                 'value': 0},
        'LEDPwrLvl-SP':         {'type': 'float', 'unit': '%',
                                 'hihi': 100, 'high': 100, 'hilim': 100,
                                 'lolim': 0, 'low': 0, 'lolo': 0},
        'LEDPwrLvl-RB':         {'type': 'float', 'unit': '%',
                                 'hihi': 100, 'high': 100, 'hilim': 100,
                                 'lolim': 0, 'low': 0, 'lolo': 0},
        'LEDPwrScaleFactor-SP': {'type': 'float', 'value': 0.011, 'prec': 6,
                                 'hihi': 0.02, 'high': 0.02, 'hilim': 0.02,
                                 'lolim': 0.000001, 'low': 0.000001,
                                 'lolo': 0.000001},
        'LEDPwrScaleFactor-RB': {'type': 'float', 'value': 0.011, 'prec': 6,
                                 'hihi': 0.02, 'high': 0.02, 'hilim': 0.02,
                                 'lolim': 0.000001, 'low': 0.000001,
                                 'lolo': 0.000001},
        'LEDThold-SP':          {'type': 'float', 'unit': 'V', 'prec': 4,
                                 'hihi': 10, 'high': 10, 'hilim': 10,
                                 'lolim': 0, 'low': 0, 'lolo': 0},
        'LEDThold-RB':          {'type': 'float', 'unit': 'V', 'prec': 4,
                                 'hihi': 10, 'high': 10, 'hilim': 10,
                                 'lolim': 0, 'low': 0, 'lolo': 0},

        # Camera statistics unit conversion PVs (pixels -> mm)
        'ImgScaleFactor-SP':   {'type': 'float', 'value': 1, 'prec': 6,
                                'hihi': 1000000, 'high': 1000000,
                                'hilim': 1000000, 'lolim': 0, 'low': 0,
                                'lolo': 0},
        'ImgScaleFactor-RB':   {'type': 'float', 'value': 1, 'prec': 6,
                                'hihi': 1000000, 'high': 1000000,
                                'hilim': 1000000, 'lolim': 0, 'low': 0,
                                'lolo': 0},
        'ImgCenterOffsetX-SP': {'type': 'float', 'value': 0, 'prec': 6,
                                'hihi': 1000000, 'high': 1000000,
                                'hilim': 1000000, 'lolim': -1000000,
                                'low': -1000000, 'lolo': -1000000},
        'ImgCenterOffsetX-RB': {'type': 'float', 'value': 0, 'prec': 6,
                                'hihi': 1000000, 'high': 1000000,
                                'hilim': 1000000, 'lolim': -1000000,
                                'low': -1000000, 'lolo': -1000000},
        'ImgCenterOffsetY-SP': {'type': 'float', 'value': 0, 'prec': 6,
                                'hihi': 1000000, 'high': 1000000,
                                'hilim': 1000000, 'lolim': -1000000,
                                'low': -1000000, 'lolo': -1000000},
        'ImgCenterOffsetY-RB': {'type': 'float', 'value': 0, 'prec': 6,
                                'hihi': 1000000, 'high': 1000000,
                                'hilim': 1000000, 'lolim': -1000000,
                                'low': -1000000, 'lolo': -1000000},
    }
    return pvs_database



"""."""

from ..namesys import SiriusPVName as _PVName
from .device import Device as _Device, DeviceSet as _DeviceSet


class Screen(_DeviceSet):
    """."""

    class DEVICES:
        """Devices names."""

        TB_1 = 'TB-01:DI-Scrn-1'
        TB_2 = 'TB-01:DI-Scrn-2'
        TB_3 = 'TB-02:DI-Scrn-1'
        TB_4 = 'TB-02:DI-Scrn-2'
        TB_5 = 'TB-03:DI-Scrn'
        TB_6 = 'TB-04:DI-Scrn'
        BO_1 = 'BO-01D:DI-Scrn-1'
        BO_2 = 'BO-01D:DI-Scrn-2'
        BO_3 = 'BO-02U:DI-Scrn'
        TS_1 = 'TS-01:DI-Scrn'
        TS_2 = 'TS-02:DI-Scrn'
        TS_3 = 'TS-03:DI-Scrn'
        TS_4 = 'TS-04:DI-Scrn-1'
        TS_5 = 'TS-04:DI-Scrn-2'
        TS_6 = 'TS-04:DI-Scrn-3'
        TB = (TB_1, TB_2, TB_3, TB_4, TB_5, TB_6)
        TS = (TS_1, TS_2, TS_3, TS_4, TS_5, TS_6)
        BO = (BO_1, BO_2, BO_3)
        ALL = (
            TB_1, TB_2, TB_3, TB_4, TB_5, TB_6,
            BO_1, BO_2, BO_3,
            TS_1, TS_2, TS_3, TS_4, TS_5, TS_6,
            )

    def __init__(self, devname):
        """."""
        # check if device exists
        if devname not in Screen.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        self.screen = _Screen(devname)
        self.screencam = _ScreenCam(devname)
        devs = [self.screen, self.screencam]
        super().__init__(devs, devname=devname)

    @property
    def image_width(self):
        """."""
        return self.screen['ImgROIWidth-RB']

    @property
    def image_height(self):
        """."""
        return self.screen['ImgROIHeight-RB']

    @property
    def image(self):
        """."""
        row_image = self.screen['ImgData-Mon']
        h = self.image_height
        w = self.image_width
        return row_image.reshape([h, w])

    @property
    def centerx(self):
        """."""
        return self.screen['CenterXDimFei-Mon']

    @property
    def centery(self):
        """."""
        return self.screen['CenterYDimFei-Mon']

    @property
    def sigmax(self):
        """."""
        return self.screen['SigmaXDimFei-Mon']

    @property
    def sigmay(self):
        """."""
        return self.screen['SigmaYDimFei-Mon']

    @property
    def angle(self):
        """."""
        return self.screen['ThetaDimFei-Mon']

    @property
    def scale_factor_x(self):
        """Pixel to mm"""
        return self.screencam['ScaleFactorX-RB']

    @property
    def scale_factor_y(self):
        """Pixel to mm"""
        return self.screencam['ScaleFactorY-RB']

    @property
    def center_offset_x(self):
        """."""
        return self.screencam['CenterOffsetX-RB']

    @property
    def center_offset_y(self):
        """."""
        return self.screencam['CenterOffsetY-RB']


class _Screen(_Device):
    """."""

    DEVICES = Screen.DEVICES

    PROPERTIES_DEFAULT = (
        'ImgData-Mon',
        'CenterXDimFei-Mon', 'CenterYDimFei-Mon',
        'SigmaXDimFei-Mon', 'SigmaYDimFei-Mon',
        'ThetaDimFei-Mon', 'ImgROIHeight-RB',
        'ImgROIWidth-RB',
        )

    def __init__(self, devname, props2init='all'):
        """."""
        if devname not in Screen.DEVICES.ALL:
            raise NotImplementedError(devname)
        super().__init__(devname, props2init=props2init)


class _ScreenCam(_Device):
    """."""

    DEVICES = Screen.DEVICES

    PROPERTIES_DEFAULT = (
        'ScaleFactorX-RB', 'ScaleFactorY-RB',
        'CenterOffsetX-RB', 'CenterOffsetY-RB',
        )

    def __init__(self, devname, props2init='all'):
        """."""
        if devname not in Screen.DEVICES.ALL:
            raise NotImplementedError(devname)
        devname = _PVName(devname).substitute(dev='ScrnCam')
        super().__init__(devname, props2init=props2init)



"""Define PVs, constants and properties of High Level Orbit Interlock app."""

import os as _os
import numpy as _np

from .. import csdev as _csdev
from ..util import ClassProperty as _classproperty
from ..search import BPMSearch as _BPMSearch, LLTimeSearch as _LLTimeSearch, \
    HLTimeSearch as _HLTimeSearch
from ..namesys import SiriusPVName as _PVName
from ..diagbeam.bpm.csdev import Const as _csbpm


NR_BPM = 160


# --- Enumeration Types ---

class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    STS_LBLS_BPM = (
        'Connected',
        'PosEnblSynced', 'AngEnblSynced', 'MinSumEnblSynced', 'GlobEnblSynced',
        'PosLimsSynced', 'AngLimsSynced', 'MinSumLimsSynced',
        'AcqConfigured', 'LogTrigConfig')
    STS_LBLS_TIMING = (
        'EVGConn', 'EVGIntlkEnblSynced', 'EVGConfig',
        'FoutsConn', 'FoutsConfig',
        'AFCTimingConn', 'AFCTimingConfig',
        'AFCPhysTrigsConn', 'AFCPhysTrigsConfig',
        'OrbIntlkTrigConn', 'OrbIntlkTrigStatusOK', 'OrbIntlkTrigConfig',
        'OrbIntlkRedTrigConn', 'OrbIntlkRedTrigStatusOK',
        'OrbIntlkRedTrigConfig',
        'LLRFPsMtmTrigConn', 'LLRFPsMtmTrigStatusOK', 'LLRFPsMtmTrigConfig',
        'BPMPsMtmTrigConn', 'BPMPsMtmTrigStatusOK', 'BPMPsMtmTrigConfig',
        'DCCT13C4PsMtmTrigConn', 'DCCT13C4PsMtmTrigStatusOK',
        'DCCT13C4PsMtmTrigConfig', 'DCCT14C4PsMtmTrigConn',
        'DCCT14C4PsMtmTrigStatusOK', 'DCCT14C4PsMtmTrigConfig',
        )
    STS_LBLS_LLRF = ('Connected', 'Configured')


_et = ETypes  # syntactic sugar


# --- Const class ---

class Const(_csdev.Const):
    """Const class defining Orbit Interlock constants."""

    CONV_UM_2_NM = 1e3

    IOC_PREFIX = _PVName('SI-Glob:AP-OrbIntlk')
    DEF_TIMEOUT = 10  # [s]
    DEF_TIMESLEEP = 0.1  # [s]
    DEF_TIMEWAIT = 3  # [s]

    DEF_TIME2WAIT_INTLKREARM = 30  # [s]

    HLTRIG_2_CONFIG = [
        ('SI-Fam:TI-BPM-OrbIntlk', (
            ('Src-Sel', 4),
            ('DelayRaw-SP', 0),
            ('State-Sel', 1),
            ('WidthRaw-SP', 1),
            ('Direction-Sel', 1))),
        ('SI-Fam:TI-OrbIntlkRedundancy', (
            ('Src-Sel', 0),
            ('State-Sel', 1),
            ('Polarity-Sel', 0),
            ('Log-Sel', 1))),
        ('SI-Glob:TI-LLRF-PsMtm', (
            ('Src-Sel', 5),
            ('DelayRaw-SP', 0),
            ('State-Sel', 1),
            ('WidthRaw-SP', 9369))),
        ('SI-Fam:TI-BPM-PsMtm', (
            ('Src-Sel', 5),
            ('DelayRaw-SP', 0),
            ('State-Sel', 1),
            ('WidthRaw-SP', 6))),
        ('SI-13C4:TI-DCCT-PsMtm', (
            ('Src-Sel', 0),
            ('State-Sel', 1),
            ('Polarity-Sel', 0),
            ('Log-Sel', 0))),
        ('SI-14C4:TI-DCCT-PsMtm', (
            ('Src-Sel', 0),
            ('State-Sel', 1),
            ('Polarity-Sel', 0),
            ('Log-Sel', 0))),
    ]
    FOUTSFIXED_RXENBL = {
        'CA-RaTim:TI-Fout-2': 0b01000001,
    }
    AFCTI_CONFIGS = (
        ('DevEnbl-Sel', 1),
        ('RTMPhasePropGain-SP', 100),
        ('RTMPhaseIntgGain-SP', 1),
        ('RTMFreqPropGain-SP', 1),
        ('RTMFreqIntgGain-SP', 128),
        ('RTMPhaseNavg-SP', 0),
        ('RTMPhaseDiv-SP', 0),
    )
    AFCPHYTRIG_CONFIGS = (
        ('Dir-Sel', 0),
        ('DirPol-Sel', 1),
        ('TrnLen-SP', 20),
    )
    SIBPMLOGTRIG_CONFIGS = (
        ('TRIGGER4TrnSrc-Sel', 1),
        ('TRIGGER4TrnOutSel-SP', 2),
        ('TRIGGER_PM0RcvSrc-Sel', 0),
        ('TRIGGER_PM0RcvInSel-SP', 2),
        ('TRIGGER_PM1RcvSrc-Sel', 0),
        ('TRIGGER_PM1RcvInSel-SP', 2),
        ('TRIGGER_PM6RcvSrc-Sel', 0),
        ('TRIGGER_PM6RcvInSel-SP', 2),
        ('TRIGGER_PM7RcvSrc-Sel', 0),
        ('TRIGGER_PM7RcvInSel-SP', 2),
        ('TRIGGER_PM11RcvSrc-Sel', 0),
        ('TRIGGER_PM11RcvInSel-SP', 2),
        ('TRIGGER_PM12RcvSrc-Sel', 0),
        ('TRIGGER_PM12RcvInSel-SP', 2),
        ('TRIGGER_PM14RcvSrc-Sel', 0),
        ('TRIGGER_PM14RcvInSel-SP', 2),
    )
    REDUNDANCY_TABLE = {
        'IA-10RaBPM:TI-AMCFPGAEVR': 'IA-10RaBPM:TI-EVR',
    }

    __EVG_CONFIGS = None
    __FOUTS_2_MON = None

    @_classproperty
    def EVG_CONFIGS(cls):
        """EVG configurations"""
        if cls.__EVG_CONFIGS is not None:
            return cls.__EVG_CONFIGS

        fouts = set()
        evgchans = set()
        evgrxenbl = list()
        for ch in _LLTimeSearch.get_connections_twds_evg():
            if ch.dev != 'BPM':
                continue
            if ch.sec != 'SI':
                continue
            if ch.dev == 'BPM' and ch.sub.endswith(('SA', 'SB', 'SP')):
                continue
            fch = _LLTimeSearch.get_fout_channel(ch)
            fouts.add(fch.device_name)
            evgch = _LLTimeSearch.get_evg_channel(fch)
            if evgch in evgchans:
                continue
            evgchans.add(evgch)
            evgrxenbl.append(int(evgch.propty[3:]))

        hlevts = _HLTimeSearch.get_hl_events()
        evtin0 = int(hlevts['Intlk'].strip('Evt'))
        evtin1 = int(hlevts['ItlkR'].strip('Evt'))
        evtin2 = int(hlevts['DCT13'].strip('Evt'))
        evtin3 = int(hlevts['DCT14'].strip('Evt'))
        evtout = int(hlevts['RFKll'].strip('Evt'))
        evgconfigs = [
            ('IntlkTbl0to15-Sel', 0b000000010000001),
            ('IntlkTbl16to27-Sel', 0),
            ('IntlkCtrlRepeat-Sel', 0),
            ('IntlkCtrlRepeatTime-SP', 0),
            ('IntlkEvtIn0-SP', evtin0),
            ('IntlkEvtIn1-SP', evtin1),
            ('IntlkEvtIn2-SP', evtin2),
            ('IntlkEvtIn3-SP', evtin3),
            ('IntlkEvtOut-SP', evtout),
            ('IntlkLogEnbl-SP', 0b11111111),
            ]
        evgconfigs.extend([(f'RxEnbl-SP.B{b}', 1) for b in evgrxenbl])

        cls.__FOUTS_2_MON = fouts
        cls.__EVG_CONFIGS = evgconfigs

        return cls.__EVG_CONFIGS

    @_classproperty
    def FOUTS_2_MON(cls):
        """Fouts to be monitored."""
        cls.EVG_CONFIGS
        return cls.__FOUTS_2_MON

    AcqChan = _csbpm.AcqChan
    AcqTrigTyp = _csbpm.AcqTrigTyp
    AcqRepeat = _csbpm.AcqRepeat
    AcqStates = _csbpm.AcqStates

    def __init__(self):
        """Class constructor."""
        # crates mapping
        self.crates_map = _LLTimeSearch.get_crates_mapping()

        # trigger source to fout out mapping
        self.trigsrc2fout_map = _LLTimeSearch.get_trigsrc2fout_mapping()

        # interlock redundancy table for fout outs
        self.intlkr_fouttable = {
            self.trigsrc2fout_map[k]: self.trigsrc2fout_map[v]
            for k, v in self.REDUNDANCY_TABLE.items()}
        self.intlkr_fouttable.update(
            {v: k for k, v in self.intlkr_fouttable.items()})

        # bpm names and nicknames
        self.bpm_names = _BPMSearch.get_names({'sec': 'SI', 'dev': 'BPM'})
        if NR_BPM != len(self.bpm_names):
            raise ValueError('Inconsistent NR_BPM parameter!')
        self.bpm_nicknames = _BPMSearch.get_nicknames(self.bpm_names)

        # bpm position along the ring
        self.bpm_pos = _BPMSearch.get_positions(self.bpm_names)

        # bpm distance for each BPM pair
        aux_pos = _np.roll(self.bpm_pos, 1)
        bpm_dist = _np.diff(aux_pos)[::2]
        # copy same distance from next high beta section to injection section
        bpm_dist[0] = bpm_dist[4*4]
        bpm_dist = _np.repeat(bpm_dist, 2)
        self.bpm_dist = _np.roll(bpm_dist, -1)

        # bpm number
        self.nr_bpms = len(self.bpm_names)

        # data folder
        path = _os.path.join(
            '/home', 'sirius', 'iocs-log', 'si-ap-orbintlk', 'data')
        self.pos_enbl_fname = _os.path.join(path, 'pos_enbllist.enbl')
        self.ang_enbl_fname = _os.path.join(path, 'ang_enbllist.enbl')
        self.minsum_enbl_fname = _os.path.join(path, 'minsum_enbllist.enbl')

        self.pos_x_min_lim_fname = _os.path.join(path, 'pos_x_min.lim')
        self.pos_x_max_lim_fname = _os.path.join(path, 'pos_x_max.lim')
        self.pos_y_min_lim_fname = _os.path.join(path, 'pos_y_min.lim')
        self.pos_y_max_lim_fname = _os.path.join(path, 'pos_y_max.lim')
        self.ang_x_min_lim_fname = _os.path.join(path, 'ang_x_min.lim')
        self.ang_x_max_lim_fname = _os.path.join(path, 'ang_x_max.lim')
        self.ang_y_min_lim_fname = _os.path.join(path, 'ang_y_min.lim')
        self.ang_y_max_lim_fname = _os.path.join(path, 'ang_y_max.lim')
        self.minsum_lim_fname = _os.path.join(path, 'minsum.lim')

    def get_database(self):
        """Return Soft IOC database."""
        pvs_database = {
            # Global
            'Version-Cte': {'type': 'string', 'value': 'UNDEF'},
            'Log-Mon': {'type': 'string', 'value': 'Starting...'},

            'Enable-Sel': {
                'type': 'enum', 'enums': _et.DSBL_ENBL,
                'value': self.DsblEnbl.Dsbl},
            'Enable-Sts': {
                'type': 'enum', 'enums': _et.DSBL_ENBL,
                'value': self.DsblEnbl.Dsbl},
            'BPMStatus-Mon': {'type': 'int', 'value': 0b111111111},
            'TimingStatus-Mon': {'type': 'int', 'value': (1 << 19) - 1},
            'LLRFStatus-Mon': {'type': 'int', 'value': 0b11},
            'BPMStatusLabels-Cte': {
                'type': 'string', 'count': len(_et.STS_LBLS_BPM),
                'value': _et.STS_LBLS_BPM},
            'TimingStatusLabels-Cte': {
                'type': 'string', 'count': len(_et.STS_LBLS_TIMING),
                'value': _et.STS_LBLS_TIMING},
            'LLRFStatusLabels-Cte': {
                'type': 'string', 'count': len(_et.STS_LBLS_LLRF),
                'value': _et.STS_LBLS_LLRF},
            'BPMMonitoredDevices-Mon': {
                'type': 'char', 'count': 1000, 'value': ''},
            'TimingMonitoredDevices-Mon': {
                'type': 'char', 'count': 1000, 'value': ''},

            # Enable lists
            'PosEnblList-SP': {
                'type': 'int', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'BPM used in orbit position interlock'},
            'PosEnblList-RB': {
                'type': 'int', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'BPM used in orbit position interlock'},

            'AngEnblList-SP': {
                'type': 'int', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'BPM used in orbit angle interlock'},
            'AngEnblList-RB': {
                'type': 'int', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'BPM used in orbit angle interlock'},

            'MinSumEnblList-SP': {
                'type': 'int', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'BPM used with minimum sum threshold enabled'},
            'MinSumEnblList-RB': {
                'type': 'int', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'BPM used with minimum sum threshold enabled'},

            # Limits
            'PosXMinLim-SP': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'position minimum limits for X'},
            'PosXMinLim-RB': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'position minimum limits for X'},
            'PosXMaxLim-SP': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'position maximum limits for X'},
            'PosXMaxLim-RB': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'position maximum limits for X'},

            'PosYMinLim-SP': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'position minimum limits for Y'},
            'PosYMinLim-RB': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'position minimum limits for Y'},
            'PosYMaxLim-SP': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'position maximum limits for Y'},
            'PosYMaxLim-RB': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'position maximum limits for Y'},

            'AngXMinLim-SP': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'angle minimum limits for X'},
            'AngXMinLim-RB': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'angle minimum limits for X'},
            'AngXMaxLim-SP': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'angle maximum limits for X'},
            'AngXMaxLim-RB': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'angle maximum limits for X'},

            'AngYMinLim-SP': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'angle minimum limits for Y'},
            'AngYMinLim-RB': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'angle minimum limits for Y'},
            'AngYMaxLim-SP': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'angle maximum limits for Y'},
            'AngYMaxLim-RB': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
                'unit': 'angle maximum limits for Y'},

            'MinSumLim-SP': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0], 'unit': 'minimum sum limits'},
            'MinSumLim-RB': {
                'type': 'float', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0], 'unit': 'minimum sum limits'},

            # Reset
            'ResetBPMGen-Cmd': {'type': 'int', 'value': 0},
            'ResetBPMPos-Cmd': {'type': 'int', 'value': 0},
            'ResetBPMAng-Cmd': {'type': 'int', 'value': 0},
            'ResetBPM-Cmd': {'type': 'int', 'value': 0},
            'Reset-Cmd': {'type': 'int', 'value': 0},
            'ResetTimingLockLatches-Cmd': {'type': 'int', 'value': 0},
            'ResetAFCTimingRTMClk-Cmd': {'type': 'int', 'value': 0},

            # Acquisition
            'PsMtmAcqChannel-Sel': {
                'type': 'enum', 'value': self.AcqChan.TbT,
                'enums': self.AcqChan._fields},
            'PsMtmAcqChannel-Sts': {
                'type': 'enum', 'value': self.AcqChan.TbT,
                'enums': self.AcqChan._fields},
            'PsMtmAcqSamplesPre-SP': {
                'type': 'int', 'value': 20000, 'lolim': 0, 'hilim': 1_000_000},
            'PsMtmAcqSamplesPre-RB': {
                'type': 'int', 'value': 20000, 'lolim': 0, 'hilim': 1_000_000},
            'PsMtmAcqSamplesPost-SP': {
                'type': 'int', 'value': 20000, 'lolim': 0, 'hilim': 1_000_000},
            'PsMtmAcqSamplesPost-RB': {
                'type': 'int', 'value': 20000, 'lolim': 0, 'hilim': 1_000_000},
            'PsMtmAcqConfig-Cmd': {'type': 'int', 'value': 0},

            # Config devices
            'ConfigEVG-Cmd': {'type': 'int', 'value': 0},
            'ConfigFouts-Cmd': {'type': 'int', 'value': 0},
            'ConfigAFCTiming-Cmd': {'type': 'int', 'value': 0},
            'ConfigHLTriggers-Cmd': {'type': 'int', 'value': 0},
            'ConfigLLRFIntlk-Cmd': {'type': 'int', 'value': 0},
            'ConfigBPMs-Cmd': {'type': 'int', 'value': 0},
            'ConfigAFCPhyTrigs-Cmd': {'type': 'int', 'value': 0},
        }
        pvs_database = _csdev.add_pvslist_cte(pvs_database)
        return pvs_database



"""Define PVs, contants and properties of PosAng SoftIOCs."""

from .. import csdev as _csdev


# --- Enumeration Types ---

class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    NEED_REF_UPDATE = ('Ok', 'NeedUpdate')


_et = ETypes  # syntactic sugar


# --- Const class ---


class Const:
    """Const class defining PosAng constants."""

    TB_CORRH_POSANG = ('TB-04:PS-CH-1', 'TB-04:PU-InjSept')
    TB_CORRV_POSANG = ('TB-04:PS-CV-1', 'TB-04:PS-CV-2')

    TS_CORRH_POSANG_CHSEPT = ('TS-04:PS-CH', 'TS-04:PU-InjSeptF')
    TS_CORRH_POSANG_SEPTSEPT = \
        ('TS-04:PU-InjSeptG-1', 'TS-04:PU-InjSeptG-2', 'TS-04:PU-InjSeptF')
    TS_CORRV_POSANG = \
        ('TS-04:PS-CV-0', 'TS-04:PS-CV-1', 'TS-04:PS-CV-1E2', 'TS-04:PS-CV-2')

    STATUSLABELS = ('PS Connection', 'PS PwrState', 'PS OpMode', 'PS CtrlMode')

    NeedRefUpdate = _csdev.Const.register('NeedRefUpdate', _et.NEED_REF_UPDATE)


_c = Const  # syntactic sugar


def get_posang_database(tl, corrs_type='ch-sept'):
    """Return Soft IOC database."""
    if tl.upper() == 'TS':
        if corrs_type == 'ch-sept':
            corrh = _c.TS_CORRH_POSANG_CHSEPT
        else:
            corrh = _c.TS_CORRH_POSANG_SEPTSEPT
        corrv = _c.TS_CORRV_POSANG
        ch1_kick_unit = 'mrad'
    elif tl.upper() == 'TB':
        corrh = _c.TB_CORRH_POSANG
        corrv = _c.TB_CORRV_POSANG
        ch1_kick_unit = 'urad'

    pvs_database = {
        'Version-Cte':       {'type': 'string', 'value': 'UNDEF'},

        'Log-Mon':           {'type': 'string', 'value': 'Starting...'},

        'DeltaPosX-SP':      {'type': 'float', 'value': 0, 'prec': 6,
                              'unit': 'mm', 'hilim': 10, 'lolim': -10},
        'DeltaPosX-RB':      {'type': 'float', 'value': 0, 'prec': 6,
                              'unit': 'mm'},
        'DeltaAngX-SP':      {'type': 'float', 'value': 0, 'prec': 6,
                              'unit': 'mrad', 'hilim': 10, 'lolim': -10},
        'DeltaAngX-RB':      {'type': 'float', 'value': 0, 'prec': 6,
                              'unit': 'mrad'},

        'DeltaPosY-SP':      {'type': 'float', 'value': 0, 'prec': 6,
                              'unit': 'mm', 'hilim': 10, 'lolim': -10},
        'DeltaPosY-RB':      {'type': 'float', 'value': 0, 'prec': 6,
                              'unit': 'mm'},
        'DeltaAngY-SP':      {'type': 'float', 'value': 0, 'prec': 6,
                              'unit': 'mrad', 'hilim': 10, 'lolim': -10},
        'DeltaAngY-RB':      {'type': 'float', 'value': 0, 'prec': 6,
                              'unit': 'mrad'},

        'ConfigName-SP':     {'type': 'string', 'value': ''},
        'ConfigName-RB':     {'type': 'string', 'value': ''},
        'RespMatX-Mon':      {'type': 'float', 'value': 4*[0], 'prec': 6,
                              'count': 4},
        'RespMatY-Mon':      {'type': 'float', 'value': 4*[0], 'prec': 6,
                              'count': 4},

        'CH1-Cte':           {'type': 'string', 'value': corrh[0]},
        'RefKickCH1-Mon':    {'type': 'float', 'value': 0, 'prec': 4,
                              'unit': ch1_kick_unit},
        'CH2-Cte':           {'type': 'string', 'value': corrh[1]},
        'RefKickCH2-Mon':    {'type': 'float', 'value': 0, 'prec': 4,
                              'unit': 'mrad'},
        'CV1-Cte':           {'type': 'string', 'value':  corrv[0]},
        'RefKickCV1-Mon':    {'type': 'float', 'value': 0, 'prec': 4,
                              'unit': 'urad'},
        'CV2-Cte':           {'type': 'string', 'value':  corrv[1]},
        'RefKickCV2-Mon':    {'type': 'float', 'value': 0, 'prec': 4,
                              'unit': 'urad'},
        'SetNewRefKick-Cmd': {'type': 'int', 'value': 0},
        'NeedRefUpdate-Mon': {'type': 'enum', 'enums': _et.NEED_REF_UPDATE,
                              'value': _c.NeedRefUpdate.NeedUpdate},

        'ConfigPS-Cmd':      {'type': 'int', 'value': 0},

        'Status-Mon':        {'type': 'int', 'value': 0b1111},
        'StatusLabels-Cte':  {'type': 'char', 'count': 1000,
                              'value': '\n'.join(_c.STATUSLABELS)},
    }
    if len(corrh) == 3:
        pvs_database['CH3-Cte'] = {'type': 'string', 'value': corrh[2]}
        pvs_database['RefKickCH3-Mon'] = {'type': 'float', 'value': 0,
                                          'prec': 4, 'unit': 'mrad'}
    if len(corrv) == 4:
        pvs_database['CV3-Cte'] = {'type': 'string', 'value': corrv[2]}
        pvs_database['RefKickCV3-Mon'] = {'type': 'float', 'value': 0,
                                          'prec': 4, 'unit': 'urad'}
        pvs_database['CV4-Cte'] = {'type': 'string', 'value': corrv[3]}
        pvs_database['RefKickCV4-Mon'] = {'type': 'float', 'value': 0,
                                          'prec': 4, 'unit': 'urad'}
    pvs_database = _csdev.add_pvslist_cte(pvs_database)
    return pvs_database



"""Power Supply Controller subpackage.

This subpackage contains higher level modules and classes that are responsible
for communicating power supplies.

Modules:

    pscontrollers.py :
        Power supply classes that convert epics properties to PRUController
        method calls.

    psmodel.py :
        Auxiliary classes that help creation of power controllers by
        storing psmodel-specific constants and structures.

    pscreaders.py :
        PSController reader classes, reciped to convert epics read properties
        to PRUCrontroller method calls.

    pscwriters.py :
        PSController reader classes, reciped to convert epics write properties
        to PRUCrontroller method calls.

    pscstatus.py :
        Auxiliary class to help convert PS_STATUS bsmp variable into
        PwrState and OpMode epics properties, and vice-versa.

"""



#!/usr/bin/env python-sirius

"""Unittest module for posang.py."""

from unittest import TestCase
from siriuspy import util
from siriuspy.posang import csdev
from siriuspy.posang.csdev import get_posang_database


PUB_INTERFACE = (
        'ETypes',
        'Const',
        'get_posang_database',
    )


class TestPosAngCSDevice(TestCase):
    """Test siriuspy.posanf.csdev module."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(csdev, PUB_INTERFACE)
        self.assertTrue(valid)

    def test_get_posang_database(self):
        """Test get_posang_database."""
        dbase = get_posang_database('TB')
        self.assertIsInstance(dbase, dict)

        # PV names
        self.assertTrue('Version-Cte' in dbase)
        self.assertTrue('Log-Mon' in dbase)
        self.assertTrue('DeltaPosX-SP' in dbase)
        self.assertTrue('DeltaPosX-RB' in dbase)
        self.assertTrue('DeltaAngX-SP' in dbase)
        self.assertTrue('DeltaAngX-RB' in dbase)
        self.assertTrue('DeltaPosY-SP' in dbase)
        self.assertTrue('DeltaPosY-RB' in dbase)
        self.assertTrue('DeltaAngY-SP' in dbase)
        self.assertTrue('DeltaAngY-RB' in dbase)
        self.assertTrue('ConfigName-SP' in dbase)
        self.assertTrue('ConfigName-RB' in dbase)
        self.assertTrue('RespMatX-Mon' in dbase)
        self.assertTrue('RespMatY-Mon' in dbase)
        self.assertTrue('CH1-Cte' in dbase)
        self.assertTrue('CH2-Cte' in dbase)
        self.assertTrue('CV1-Cte' in dbase)
        self.assertTrue('CV2-Cte' in dbase)
        self.assertTrue('RefKickCH1-Mon' in dbase)
        self.assertTrue('RefKickCH2-Mon' in dbase)
        self.assertTrue('RefKickCV1-Mon' in dbase)
        self.assertTrue('RefKickCV2-Mon' in dbase)
        self.assertTrue('SetNewRefKick-Cmd' in dbase)
        self.assertTrue('NeedRefUpdate-Mon' in dbase)
        self.assertTrue('ConfigPS-Cmd' in dbase)
        self.assertTrue('Status-Mon' in dbase)
        self.assertTrue('StatusLabels-Cte' in dbase)

        # PVs units
        self.assertEqual(dbase['DeltaPosX-SP']['unit'], 'mm')
        self.assertEqual(dbase['DeltaPosX-RB']['unit'], 'mm')
        self.assertEqual(dbase['DeltaAngX-SP']['unit'], 'mrad')
        self.assertEqual(dbase['DeltaAngX-RB']['unit'], 'mrad')
        self.assertEqual(dbase['DeltaPosY-SP']['unit'], 'mm')
        self.assertEqual(dbase['DeltaPosY-RB']['unit'], 'mm')
        self.assertEqual(dbase['DeltaAngY-SP']['unit'], 'mrad')
        self.assertEqual(dbase['DeltaAngY-RB']['unit'], 'mrad')
        self.assertEqual(dbase['RefKickCH1-Mon']['unit'], 'urad')
        self.assertEqual(dbase['RefKickCH2-Mon']['unit'], 'mrad')
        self.assertEqual(dbase['RefKickCV1-Mon']['unit'], 'urad')
        self.assertEqual(dbase['RefKickCV2-Mon']['unit'], 'urad')



"""Tune devices."""


from .device import Device as _Device
from .device import DeviceSet as _DeviceSet


class TuneFrac(_Device):
    """Tune Frac device."""

    DEF_TIMEOUT = 1  # [s]

    class DEVICES:
        """Devices names."""

        SI_H = 'SI-Glob:DI-Tune-H'
        SI_V = 'SI-Glob:DI-Tune-V'
        ALL = (SI_H, SI_V)

    PROPERTIES_DEFAULT = ('TuneFrac-Mon', 'Enbl-Sel', 'Enbl-Sts')

    def __init__(self, devname, props2init='all'):
        """Init."""
        if devname not in TuneFrac.DEVICES.ALL:
            raise NotImplementedError(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def tune(self):
        """Tune Frac."""
        return self['TuneFrac-Mon']

    @property
    def enable(self):
        """Enable status."""
        return self['Enbl-Sts']

    def cmd_enable(self, timeout=DEF_TIMEOUT):
        """Enable."""
        self['Enbl-Sel'] = 1
        return self._wait('Enbl-Sts', value=1, timeout=timeout)

    def cmd_disable(self, timeout=DEF_TIMEOUT):
        """Disable."""
        self['Enbl-Sel'] = 0
        return self._wait('Enbl-Sts', value=0, timeout=timeout)


class TuneProc(_Device):
    """Tune Proc device."""

    class DEVICES:
        """Devices names."""

        SI_H = 'SI-Glob:DI-TuneProc-H'
        SI_V = 'SI-Glob:DI-TuneProc-V'
        ALL = (SI_H, SI_V)

    PROPERTIES_DEFAULT = ('Trace-Mon', )

    def __init__(self, devname, props2init='all'):
        """Init."""
        # check if device exists
        if devname not in TuneProc.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, props2init=props2init)

    @property
    def tune_wfm(self):
        """Tune waveform."""
        return self['Trace-Mon']


class Tune(_DeviceSet):
    """Tune device."""

    class DEVICES:
        """Devices names."""

        SI = 'SI-Glob:DI-Tune'
        ALL = (SI, )

    def __init__(self, devname=None):
        """Init."""
        # check if device exists
        if devname is None:
            devname = Tune.DEVICES.SI
        if devname not in Tune.DEVICES.ALL:
            raise NotImplementedError(devname)

        tune_frac_h = TuneFrac(TuneFrac.DEVICES.SI_H)
        tune_frac_v = TuneFrac(TuneFrac.DEVICES.SI_V)
        tune_proc_h = TuneProc(TuneProc.DEVICES.SI_H)
        tune_proc_v = TuneProc(TuneProc.DEVICES.SI_V)

        devices = (tune_frac_h, tune_frac_v, tune_proc_h, tune_proc_v)

        # call base class constructor
        super().__init__(devices, devname=devname)

    @property
    def tunex(self):
        """Tune Frac X."""
        return self.devices[0].tune

    @property
    def tuney(self):
        """Tune Frac Y."""
        return self.devices[1].tune

    @property
    def tunex_wfm(self):
        """Tune waveform X."""
        return self.devices[2].tune_wfm

    @property
    def tuney_wfm(self):
        """Tune waveform Y."""
        return self.devices[3].tune_wfm

    @property
    def enablex(self):
        """Tune X enable status."""
        return self.devices[0].enable

    @property
    def enabley(self):
        """Tune Y enable status."""
        return self.devices[1].enable

    def cmd_enablex(self, timeout=TuneFrac.DEF_TIMEOUT):
        """Enable tune X."""
        return self.devices[0].cmd_enable(timeout=timeout)

    def cmd_enabley(self, timeout=TuneFrac.DEF_TIMEOUT):
        """Enable tune Y."""
        return self.devices[1].cmd_enable(timeout=timeout)

    def cmd_disablex(self, timeout=TuneFrac.DEF_TIMEOUT):
        """Disable tune X."""
        return self.devices[0].cmd_disable(timeout=timeout)

    def cmd_disabley(self, timeout=TuneFrac.DEF_TIMEOUT):
        """Disable tune Y."""
        return self.devices[1].cmd_disable(timeout=timeout)


class TuneCorr(_Device):
    """TuneCorr device."""

    class DEVICES:
        """Devices names."""

        SI = 'SI-Glob:AP-TuneCorr'
        ALL = (SI, )

    PROPERTIES_DEFAULT = (
        'DeltaTuneX-SP', 'DeltaTuneX-RB',
        'DeltaTuneY-SP', 'DeltaTuneY-RB',
        'SetNewRefKL-Cmd', 'ApplyDelta-Cmd',
        )

    def __init__(self, devname=None, props2init='all'):
        """Init."""
        if devname is None:
            devname = TuneCorr.DEVICES.SI
        if devname not in TuneCorr.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, props2init=props2init)

    @property
    def delta_tunex(self):
        """Delta Tune X."""
        return self['DeltaTuneX-RB']

    @delta_tunex.setter
    def delta_tunex(self, value):
        self['DeltaTuneX-SP'] = value

    @property
    def delta_tuney(self):
        """Delta Tune Y."""
        return self['DeltaTuneY-RB']

    @delta_tuney.setter
    def delta_tuney(self, value):
        self['DeltaTuneY-SP'] = value

    def cmd_update_reference(self):
        """Update reference tunes."""
        self['SetNewRefKL-Cmd'] = 1
        return True

    def cmd_apply_delta(self):
        """Apply delta tunes."""
        self['ApplyDelta-Cmd'] = 1
        return True



"""."""
from .orbit import EpicsOrbit
from .matrix import EpicsMatrix
from .correctors import EpicsCorrectors
from .csdev import SOFBFactory
from .main import SOFB

del orbit, matrix, correctors, csdev, main

__all__ = ('correctors', 'main', 'matrix', 'orbit', 'bpms', 'csdev', 'utils')
